/*
 * Copyright NetApp Inc, 2021 All rights reserved
 */

package grafana

import (
	"bytes"
	"crypto/tls"
	"encoding/json"
	"errors"
	"fmt"
	"github.com/netapp/harvest/v2/cmd/harvest/version"
	"github.com/netapp/harvest/v2/pkg/conf"
	"github.com/netapp/harvest/v2/pkg/requests"
	"github.com/netapp/harvest/v2/pkg/util"
	goversion "github.com/netapp/harvest/v2/third_party/go-version"
	"github.com/netapp/harvest/v2/third_party/tidwall/sjson"
	"github.com/spf13/cobra"
	"github.com/tidwall/gjson"
	"golang.org/x/text/cases"
	"golang.org/x/text/language"
	"io"
	"net/http"
	"os"
	"path/filepath"
	"regexp"
	"strconv"
	"strings"
	"time"
)

const (
	clientTimeout                 = 5
	DefaultDataSource             = "prometheus"
	GPerm             os.FileMode = 0644
)

var (
	grafanaMinVers = "7.1.0" // lowest grafana version we require
	homePath       string
	harvestRelease = version.VERSION
)

type options struct {
	command             string // one of: import, export, clean
	addr                string // URL of Grafana server (e.g. "http://localhost:3000")
	token               string // API token issued by Grafana server
	dir                 string
	serverfolder        Folder
	datasource          string
	variable            bool
	forceImport         bool
	client              *http.Client
	headers             http.Header
	config              string
	prefix              string
	useHTTPS            bool
	useInsecureTLS      bool
	overwrite           bool
	labels              []string
	dirGrafanaFolderMap map[string]*Folder
	addMultiSelect      bool
	svmRegex            string
	customizeDir        string
}

type Folder struct {
	name string // Grafana folder where to upload from where to download dashboards
	id   int64
	uid  string
}

func adjustOptions() {
	opts.config = conf.ConfigPath(opts.config)
	homePath = conf.Path("")
	opts.dirGrafanaFolderMap = make(map[string]*Folder)

	// When opt.addr starts with https don't change it
	if !strings.HasPrefix(opts.addr, "https://") {
		//goland:noinspection HttpUrlsUsage
		opts.addr = strings.TrimPrefix(opts.addr, "http://")
		opts.addr = strings.TrimPrefix(opts.addr, "https://")
		opts.addr = strings.TrimSuffix(opts.addr, "/")
		if opts.useHTTPS {
			opts.addr = "https://" + opts.addr
		} else {
			//goland:noinspection HttpUrlsUsage
			opts.addr = "http://" + opts.addr
		}
	}
}

func askForToken() {
	// ask for API token if not provided as arg and validate
	if err := checkToken(opts, false, 5); err != nil {
		fmt.Println(err)
		os.Exit(1)
	}
}

func doCustomize(_ *cobra.Command, _ []string) {
	adjustOptions()
	exitIfExist(opts.customizeDir, "output-dir")

	initImportVars()
	importDashboards(opts)
}

func doExport(_ *cobra.Command, _ []string) {
	adjustOptions()
	exitIfExist(opts.dir, "directory")
	askForToken()
	initExportVars()
	fmt.Printf("preparing to export dashboards from serverfolder %s directory %s\n", opts.serverfolder.name, opts.dir)
	exportDashboards(opts)
}

func initExportVars() {
	opts.dirGrafanaFolderMap[opts.dir] = &Folder{name: opts.serverfolder.name}
}

func exportDashboards(opts *options) {
	for k, v := range opts.dirGrafanaFolderMap {
		err := exportFiles(k, v)
		if err != nil {
			fmt.Printf("Error during export %v\n", err)
		}
	}
}

func exportFiles(dir string, folder *Folder) error {
	var (
		err   error
		uids  map[string]string
		count int
	)

	err = checkFolder(folder)
	if err != nil {
		fmt.Printf("folder %s error %v\n", folder.name, err)
		os.Exit(1)
	}
	if folder.id == 0 {
		fmt.Printf("error folder %s doesn't exist in grafana, unable to continue\n", folder.name)
		os.Exit(1)
	}
	fmt.Printf("querying for content of folder name [%s]\n", folder.name)

	result, status, code, err := sendRequestArray(opts, "GET", "/api/search?folderIds="+strconv.FormatInt(folder.id, 10), nil)
	if err != nil && code != 200 {
		fmt.Printf("server response [%d: %s]: %v\n", code, status, err)
		return err
	}

	uids = make(map[string]string)
	rep := strings.NewReplacer("/", "_", "-", "_")
	for _, elem := range result {
		uid := elem["uid"].(string)
		uri := elem["uri"].(string)
		if uid != "" && uri != "" {
			uids[uid] = rep.Replace(uri)
		}
	}

	if err = os.MkdirAll(dir, 0750); err != nil {
		fmt.Printf("error makedir [%s]: %v\n", dir, err)
		return err
	}
	fmt.Printf("exporting dashboards to directory [%s]\n", dir)

	for uid, uri := range uids {
		result, status, code, err := sendRequest(opts, "GET", "/api/dashboards/uid/"+uid, nil)
		if err != nil {
			fmt.Printf("error requesting [%s]: [%d: %s] %v\n", uid, code, status, err)
			return err
		}
		if dashboard, ok := result["dashboard"]; ok {
			fp := filepath.Join(dir, uri+".json")
			data, err := json.Marshal(dashboard)
			if err != nil {
				fmt.Printf("error marshall dashboard [%s]: %v\n\n", uid, err)
				return err
			}
			// creating dashboards with group and other permissions of read are OK
			if err = os.WriteFile(fp, data, GPerm); err != nil {
				fmt.Printf("error write to [%s]: %v\n", fp, err)
				return err
			}
			fmt.Printf("OK - exported [%s]\n", fp)
			count++
		}
	}
	fmt.Printf("exported %d dashboards to [%s]\n", count, dir)
	return nil
}

func addSvmRegex(content []byte, fileName string, val string) []byte {
	svmExpression := []string{"templating.list.#(name=\"SVM\")"}
	if fileName == "snapmirror.json" {
		svmExpression = []string{"templating.list.#(name=\"DestinationSVM\")", "templating.list.#(name=\"SourceSVM\")"}
	}
	for _, s := range svmExpression {
		var err error
		svm := gjson.GetBytes(content, s)
		if svm.Exists() {
			content, err = sjson.SetBytes(content, s+".regex", []byte(val))
			if err != nil {
				fmt.Printf("Error while setting svm regex: %v\n", err)
				continue
			}
			content, err = sjson.SetBytes(content, s+".allValue", json.RawMessage("null"))
			if err != nil {
				fmt.Printf("Error while setting svm allValue: %v\n", err)
				continue
			}
		}
	}
	return content
}

func addLabel(content []byte, label string, labelMap map[string]string) []byte {
	// extract the list of variables
	templateList := gjson.GetBytes(content, "templating.list")
	if !templateList.Exists() {
		fmt.Printf("No template variables found, ignoring add label")
		return content
	}
	vars := templateList.Array()

	// create a new list of vars and copy the existing ones into it, duplicate the first var since we're going to
	// overwrite it
	var newVars []gjson.Result
	newVars = append(newVars, vars[0])
	newVars = append(newVars, vars...)

	// Assume Datasource is first and insert the new label var as the second element.
	// If we're wrong, that's OK, no harm
	newVars[1] = gjson.ParseBytes(newLabelVar(label))

	// Write the variables into a string builder
	// Modify the existing variables by adding the new label query
	varsString := strings.Builder{}
	varsString.WriteString("[")
	for i, result := range newVars {
		aStr := addChainedVar(result, label, labelMap)
		if aStr == "" {
			varsString.WriteString(result.String())
		} else {
			varsString.WriteString(aStr)
		}
		if i < len(newVars)-1 {
			varsString.WriteString(",\n")
		}
	}
	varsString.WriteString("]")

	newContent, err := sjson.SetRawBytes(content, "templating.list", []byte(varsString.String()))
	if err != nil {
		fmt.Printf("error inserting label=[%s] into dashboard, err: %+v", label, err)
		return content
	}
	return newContent
}

func addChainedVar(result gjson.Result, label string, labelMap map[string]string) string {
	varName := result.Get("name")
	definition := result.Get("definition")
	query := result.Get("query.query")

	if !varName.Exists() || !definition.Exists() || !query.Exists() {
		return ""
	}
	// Don't modify the query if this is one of the new labels we're adding since its query is already correct
	if _, ok := labelMap[varName.String()]; ok {
		return ""
	}

	defStr := definition.String()
	if defStr != query.String() {
		return ""
	}
	chained := toChainedVar(defStr, label)
	if chained == "" {
		return ""
	}
	rString := result.String()
	var err error
	rString, err = sjson.Set(rString, "definition", chained)
	if err != nil {
		fmt.Printf("error setting definition of varName=[%s] for label=[%s], err: %+v", varName, label, err)
		return ""
	}
	rString, err = sjson.Set(rString, "query.query", chained)
	if err != nil {
		fmt.Printf("error setting query of varName=[%s] for label=[%s], err: %+v", varName, label, err)
		return ""
	}
	return rString
}

func toChainedVar(defStr string, label string) string {
	if !strings.Contains(defStr, "label_values") {
		return ""
	}

	title := cases.Title(language.Und).String(label)
	lastBracket := strings.LastIndex(defStr, "}")
	if lastBracket == -1 {
		lastParen := strings.LastIndex(defStr, ")")
		if lastParen == -1 {
			return ""
		}

		lastComma := strings.LastIndex(defStr, ",")
		firstParen := strings.Index(defStr, "(")
		if firstParen == -1 {
			return ""
		}
		if lastComma == -1 { //nolint:revive
			// Case 1: There are not existing labels
			// label_values(datacenter) becomes label_values({foo=~"$Foo"}, datacenter)
		} else {
			// Case 2: There is a single metric
			// label_values(poller_status, datacenter) becomes label_values(poller_status{org=~"$org"}, datacenter)
			return defStr[0:lastComma] + "{" + label + `=~"$` + title + `"},` + defStr[lastComma+1:]
		}
		if firstParen+1 > len(defStr) {
			return ""
		}
		return defStr[0:firstParen] + "({" + label + `=~"$` + title + `"}, ` + defStr[firstParen+1:]
	}
	// Case 2: There are existing metrics
	// label_values(aggr_new_status{datacenter="$Datacenter",cluster="$Cluster"}, node) becomes
	// label_values(aggr_new_status{datacenter="$Datacenter",cluster="$Cluster",foo=~"$Foo"}, node)
	return defStr[0:lastBracket] + "," + label + `=~"$` + title + `"` + defStr[lastBracket:]
}

func newLabelVar(label string) []byte {
	return []byte(fmt.Sprintf(`{
  "allValue": ".*",
  "current": {
    "selected": false
  },
  "definition": "label_values(%s)",
  "hide": 0,
  "includeAll": true,
  "multi": true,
  "name": "%s",
  "options": [],
  "query": {
    "query": "label_values(%s)",
    "refId": "StandardVariableQuery"
  },
  "refresh": 2,
  "regex": "",
  "skipUrlSync": false,
  "sort": 0,
  "type": "query"
}`, label, cases.Title(language.Und).String(label), label))
}

func doImport(_ *cobra.Command, _ []string) {
	opts.command = "import"
	_, err := conf.LoadHarvestConfig(opts.config)
	if err != nil {
		return
	}

	adjustOptions()
	validateImport()
	askForToken()
	initImportVars()

	fmt.Printf("preparing to import dashboards...\n")
	importDashboards(opts)
}

func validateImport() {
	// default case
	if opts.dir == "" && opts.serverfolder.name == "" {
		opts.dir = "grafana/dashboards"
		opts.dir = filepath.Join(homePath, opts.dir)
	}

	exitIfMissing(opts.dir, "directory")
	files, err := os.ReadDir(opts.dir)
	if err != nil {
		fmt.Printf("Error %v while reading dir [%s] is the directory correct?\n", err, opts.dir)
		os.Exit(1)
	}
	if len(files) == 0 {
		fmt.Printf("No dashboards found in [%s] is the directory correct?\n", opts.dir)
		os.Exit(1)
	}
}

func initImportVars() {
	m := make(map[string]*Folder)

	// default behaviour
	if opts.dir == "grafana/dashboards" && opts.serverfolder.name == "" {
		m[filepath.Join(opts.dir, "/cmode")] = &Folder{name: "Harvest-" + harvestRelease + "-cDOT"}
		m[filepath.Join(opts.dir, "/cmode/details")] = &Folder{name: "Harvest-" + harvestRelease + "-cDOT Details"}
		m[filepath.Join(opts.dir, "/7mode")] = &Folder{name: "Harvest-" + harvestRelease + "-7mode"}
		m[filepath.Join(opts.dir, "/storagegrid")] = &Folder{name: "Harvest-" + harvestRelease + "-StorageGrid"}
	} else if opts.dir != "" && opts.serverfolder.name != "" {
		m[opts.dir] = &Folder{name: opts.serverfolder.name}
	}

	for k, v := range m {
		if opts.customizeDir == "" {
			err := checkAndCreateServerFolder(v)
			if err != nil {
				fmt.Print(err)
				os.Exit(1)
			}
		}
		opts.dirGrafanaFolderMap[k] = v
	}
}

func checkAndCreateServerFolder(folder *Folder) error {
	err := checkFolder(folder)
	if err != nil {
		fmt.Printf("error %v for folder %s\n", err, folder.name)
		os.Exit(1)
	}

	if folder.uid != "" && folder.id != 0 {
		fmt.Printf("folder [%s] exists in Grafana - OK\n", folder.name)
	} else if err = createServerFolder(folder); err != nil {
		return err
	} else {
		fmt.Printf("created Grafana folder [%s] - OK\n", folder.name)
	}
	return nil
}

func exitIfMissing(fp string, s string) {
	if _, err := os.Stat(fp); os.IsNotExist(err) {
		fmt.Printf("error: %s [%s] does not exist.\n", s, fp)
		os.Exit(1)
	}
}

func exitIfExist(fp string, s string) {
	if _, err := os.Stat(fp); err == nil {
		fmt.Printf("error: %s folder [%s] exists. Please specify an empty or non-existent directory.\n", s, fp)
		os.Exit(1)
	}
}

func importDashboards(opts *options) {
	for k, v := range opts.dirGrafanaFolderMap {
		importFiles(k, v)
	}
}

func importFiles(dir string, folder *Folder) {
	var (
		request, dashboard map[string]any
		files              []os.DirEntry
		importedFiles      int
		data               []byte
		err                error
	)
	if dir == "" {
		return
	}
	if files, err = os.ReadDir(dir); err != nil {
		// TODO check for not exist
		return
	}

	for _, file := range files {
		if !strings.HasSuffix(file.Name(), ".json") {
			continue
		}

		if data, err = os.ReadFile(filepath.Join(dir, file.Name())); err != nil {
			fmt.Printf("error reading file [%s]\n", file.Name())
			return
		}

		data = bytes.ReplaceAll(data, []byte("${DS_PROMETHEUS}"), []byte(opts.datasource))

		// If the dashboard has an uid defined, change the uid to the empty string, unless overwrite is true.
		// We do comparison for dashboard create/update based on title
		if !opts.overwrite {

			// Don't change the uid of linked dashboards since that will break the links
			isLinkedDashboard := file.Name() == "volumeBySVM.json" || file.Name() == "volumeDeepDive.json"

			if !isLinkedDashboard {
				dashboardID := gjson.GetBytes(data, "uid").String()
				if dashboardID != "" {
					data, err = sjson.SetBytes(data, "uid", []byte(""))
					if err != nil {
						fmt.Printf("error while updating the uid %s into dashboard %s, err: %+v", dashboardID, file.Name(), err)
						continue
					}
				}
			}
		}

		// If the dashboard has an id defined, change the id to empty string, unless overwrite was passed,
		// so Grafana treats this as a new dashboard instead of an update to an existing one
		if !opts.overwrite {
			if dashboardID := gjson.GetBytes(data, "id").String(); dashboardID != "" {
				data, err = sjson.SetBytes(data, "id", []byte(""))
				if err != nil {
					fmt.Printf("error while updating the id %s into dashboard %s, err: %+v", dashboardID, file.Name(), err)
					continue
				}
			}
		}

		// add svm regex
		if opts.svmRegex != "" {
			data = addSvmRegex(data, file.Name(), opts.svmRegex)
		}

		// labelMap is used to ensure we don't modify the query of one of the new labels we're adding
		labelMap := make(map[string]string)
		caser := cases.Title(language.Und)
		for _, label := range opts.labels {
			labelMap[caser.String(label)] = label
		}
		// The label is inserted in the list of variables first
		// Iterate backwards so the labels keep the same order as cmdline
		for i := len(opts.labels) - 1; i >= 0; i-- {
			data = addLabel(data, opts.labels[i], labelMap)
		}
		if err = json.Unmarshal(data, &dashboard); err != nil {
			fmt.Printf("error parsing file [%s] %+v\n", file.Name(), err)
			fmt.Println("-------------------------------")
			fmt.Println(string(data))
			fmt.Println("-------------------------------")
			return
		}

		// optionally add prefix to all metric names in the queries
		if opts.prefix != "" {
			addGlobalPrefix(dashboard, opts.prefix)
		}

		if opts.customizeDir != "" {
			err := writeCustomDashboard(dashboard, dir, file)
			if err != nil {
				fmt.Printf("error customizing dashboard [%s] %+v\n", file.Name(), err)
			}
			continue
		}

		request = make(map[string]any)
		request["overwrite"] = opts.overwrite
		request["folderId"] = folder.id
		request["dashboard"] = dashboard

		result, status, code, err := sendRequest(opts, "POST", "/api/dashboards/db", request)

		if err != nil {
			fmt.Printf("error importing [%s]  to folder [%s] \n", file.Name(), folder.name)
			return
		}

		if code != 200 {
			fmt.Printf("error importing [%s] to folder [%s] - the server replied with [%s]\n", file.Name(), folder.name, status)
			if code == 412 {
				fmt.Printf("That means you are trying to overwrite an existing dashboard.\n\n")
				fmt.Printf("If you want to overwrite, run with the --overwrite flag or choose a different Grafana folder with the --serverfolder flag.\n")
				fmt.Printf("NOTE: If the dashboard already exists, --overwrite will import the changes and increment the dashboard version number.\n\n")
			}
			fmt.Printf("Server response follows:\n")
			for k, v := range result {
				fmt.Printf("    %s => %s\n", k, v)
			}
			fmt.Println()
			return
		}
		fmt.Printf("OK - imported %s / [%s]\n", folder.name, file.Name())
		importedFiles++
	}

	if opts.customizeDir == "" {
		if importedFiles > 0 {
			fmt.Printf("Imported %d dashboards to [%s] from [%s]\n", importedFiles, folder.name, dir)
		} else {
			fmt.Printf("No dashboards found in [%s] is the directory correct?\n", dir)
		}
	}
}

func writeCustomDashboard(dashboard map[string]any, dir string, file os.DirEntry) error {
	data, err := json.Marshal(dashboard)
	if err != nil {
		return err
	}
	sub := filepath.Base(dir)
	fp := filepath.Join(opts.customizeDir, sub, file.Name())
	if err := os.MkdirAll(filepath.Dir(fp), 0750); err != nil {
		return fmt.Errorf("error makedir [%s]: %w", filepath.Dir(fp), err)
	}
	if err = os.WriteFile(fp, data, GPerm); err != nil {
		return fmt.Errorf("error writing customized dashboard to file %s: %w", fp, err)
	}
	fmt.Printf("OK - customized [%s]\n", fp)
	return nil
}

// addGlobalPrefix adds the given prefix to all metric names in the
// dashboards. It assumes that metrics are in Prometheus-format.
//
// A more reliable implementation of this feature would be, to
// add a constant prefix to all metrics, before they are pushed
// to GitHub, then replace them with a user-defined prefix
// (or empty string) when the import tool is used.
func addGlobalPrefix(db map[string]any, prefix string) {

	var (
		panels, templates, subPanels       []interface{}
		panel, templating, template, query map[string]any
		queryString, definition            string
		ok, has                            bool
	)

	// make sure prefix ends with _
	if !strings.HasSuffix(prefix, "_") {
		prefix += "_"
	}

	// apply to queries in panels
	if panels, ok = db["panels"].([]interface{}); !ok {
		return
	}

	for _, p := range panels {
		handlingPanels(p, prefix)

		// handling for sub-panels
		if panel, ok = p.(map[string]any); !ok {
			continue
		}

		if _, has = panel["panels"]; !has {
			continue
		}
		if subPanels, ok = panel["panels"].([]interface{}); ok {
			for _, subP := range subPanels {
				handlingPanels(subP, prefix)
			}
		}
	}

	// apply to variables
	if templating, ok = db["templating"].(map[string]any); !ok {
		return
	}

	if templates, ok = templating["list"].([]interface{}); !ok {
		return
	}

	for _, t := range templates {
		if template, ok = t.(map[string]any); ok {
			if definition, ok = template["definition"].(string); ok {
				template["definition"] = addPrefixToMetricNames(definition, prefix)
			}
			if query, ok = template["query"].(map[string]any); ok {
				if queryString, ok = query["query"].(string); ok {
					query["query"] = addPrefixToMetricNames(queryString, prefix)
				}
			}
		}
	}
}

func handlingPanels(p interface{}, prefix string) {
	var (
		targets       []interface{}
		panel, target map[string]any
		ok, has       bool
		expr          string
	)
	if panel, ok = p.(map[string]any); !ok {
		return
	}

	if _, has = panel["targets"]; !has {
		return
	}

	if targets, ok = panel["targets"].([]interface{}); !ok {
		return
	}

	for _, t := range targets {
		if target, ok = t.(map[string]any); !ok {
			continue
		}
		if _, has = target["expr"]; has {
			if expr, ok = target["expr"].(string); ok {
				target["expr"] = addPrefixToMetricNames(expr, prefix)
			}
		}
	}
}

// addPrefixToMetricNames adds prefix to metric names in expr or leaves it
// unchanged if no metric names are identified.
// Note that this function will only work with the Prometheus-dashboards of Harvest.
// It will use a number of patterns in which metrics might be used in queries.
// (E.g. a single metric, multiple metrics used in addition, etc. -- for examples
// see the test). If we change queries of our dashboards, we have to review
// this function as well (or come up with a better solution).
func addPrefixToMetricNames(expr, prefix string) string {
	var (
		match      [][]string
		submatch   []string
		isMatch    bool
		regex      *regexp.Regexp
		err        error
		visitedMap map[string]bool // handles if same query exist in multiple times in one expression.
	)

	// variable queries
	if strings.HasPrefix(expr, "label_values(") {
		if isMatch, err = regexp.MatchString(`^label_values\s?\(([a-zA-Z_])+(\s?{.+?})?,\s?[a-zA-Z_]+\)$`, expr); err != nil {
			fmt.Printf("Regex error: %v\n", err)
			return expr
		} else if isMatch {
			return strings.Replace(expr, "label_values(", "label_values("+prefix, 1)
		}
		return expr
	}
	// everything else is for graph queries
	regex = regexp.MustCompile(`([a-zA-Z0-9_+-]+)\s?{.+?}`)
	match = regex.FindAllStringSubmatch(expr, -1)
	visitedMap = make(map[string]bool)
	for _, m := range match {
		if _, has := visitedMap[m[1]]; !has {
			// multiple metrics used with `+`
			if strings.Contains(m[1], "+") {
				submatch = strings.Split(m[1], "+")
				for i := range submatch {
					submatch[i] = prefix + submatch[i]
				}
				expr = strings.Replace(expr, m[1], strings.Join(submatch, "+"), -1)
				// multiple metrics used with `-`
			} else if strings.Contains(m[1], "-") {
				submatch = strings.Split(m[1], "-")
				for i := range submatch {
					submatch[i] = prefix + submatch[i]
				}
				expr = strings.Replace(expr, m[1], strings.Join(submatch, "-"), -1)
				// single metric
			} else {
				expr = strings.Replace(expr, m[1], prefix+m[1], -1)
			}
			visitedMap[m[1]] = true
		}
	}

	return expr
}

func checkToken(opts *options, ignoreConfig bool, tries int) error {

	var (
		token, configPath, answer string
		err                       error
		isValidDS                 bool
	)

	if tries == 0 {
		return errors.New("no more attempts")
	}

	configPath = opts.config

	_, err = conf.LoadHarvestConfig(configPath)
	if err != nil {
		return err
	}

	if conf.Config.Tools != nil {
		if !ignoreConfig {
			token = conf.Config.Tools.GrafanaAPIToken
			fmt.Println("using API token from config")
		}
	}

	if opts.token == "" && token == "" {
		fmt.Printf("enter API token: ")
		_, _ = fmt.Scanf("%s\n", &opts.token)
	} else if opts.token == "" {
		opts.token = token
	}

	// build headers for HTTP request
	opts.headers = http.Header{}
	opts.headers.Add("Accept", "application/json")
	opts.headers.Add("Content-Type", "application/json")
	opts.headers.Add("Authorization", "Bearer "+opts.token)

	opts.client = &http.Client{Timeout: time.Duration(clientTimeout) * time.Second}
	if strings.HasPrefix(opts.addr, "https://") {
		tlsConfig := &tls.Config{InsecureSkipVerify: opts.useInsecureTLS} //nolint:gosec
		opts.client.Transport = &http.Transport{TLSClientConfig: tlsConfig}
	}
	// send random request to validate token
	result, status, code, err := sendRequest(opts, "GET", "/api/org", nil)
	if err != nil {
		return err
	} else if code != 200 && code != 404 {
		msg := result["message"].(string)
		fmt.Printf("error connect: (%d - %s) %s\n", code, status, msg)
		opts.token = ""
		return checkToken(opts, true, tries-1)
	}

	// ask user to save API key
	if conf.Config.Tools == nil || opts.token != conf.Config.Tools.GrafanaAPIToken {

		fmt.Printf("save API key for later use? [Y/n]: ")
		_, _ = fmt.Scanf("%s\n", &answer)

		if answer == "Y" || answer == "y" || answer == "yes" || answer == "" {
			if conf.Config.Tools == nil {
				conf.Config.Tools = &conf.Tools{}
			}
			conf.Config.Tools.GrafanaAPIToken = opts.token
			fmt.Printf("saving config file [%s]\n", configPath)
			if err = util.SaveConfig(configPath, opts.token); err != nil {
				return err
			}
		}
	}

	// get grafana version
	if result, _, _, err = sendRequest(opts, "GET", "/api/frontend/settings", nil); err != nil {
		return err
	}

	if opts.forceImport {
		isValidDS = true
	} else {
		isValidDS = isValidDatasource(result)
	}
	if !isValidDS {
		fmt.Printf("A Prometheus-typed datasource named \"%s\" does not exist in Grafana.\n", opts.datasource)
		os.Exit(0)
	}

	buildInfo := result["buildInfo"].(map[string]any)
	if buildInfo == nil {
		fmt.Printf("warning: unable to get grafana version. Ignoring grafana version check")
		return nil
	}
	grafanaVersion := buildInfo["version"].(string)
	if grafanaVersion == "" {
		fmt.Printf("warning: unable to get grafana version. Ignoring grafana version check")
		return nil
	}
	fmt.Printf("connected to Grafana server (version: %s)\n", grafanaVersion)
	// if we are going to import check grafana version
	if opts.command == "import" && !checkVersion(grafanaVersion) {
		fmt.Printf("warning: current set of dashboards require Grafana version (%s) or higher\n", grafanaMinVers)
		fmt.Printf("continue anyway? [y/N]: ")
		_, _ = fmt.Scanf("%s\n", &answer)
		if answer != "y" && answer != "yes" {
			os.Exit(0)
		}
	}

	return nil
}

func isValidDatasource(result map[string]any) bool {
	if result == nil {
		fmt.Printf("warning: result is null.")
		return false
	}
	datasourcesInfo := result["datasources"]
	if datasourcesInfo == nil {
		fmt.Printf("warning: datasources are missing")
		return false
	}

	for dsName, dsInfo := range datasourcesInfo.(map[string]any) {
		if dsInfo == nil {
			fmt.Printf("warning: dsInfo is missing for %s\n", dsName)
			continue
		}
		dsDetail := dsInfo.(map[string]any)
		dsType := dsDetail["type"]
		if dsType == nil {
			fmt.Printf("warning: dsType is missing for %s\n", dsName)
			continue
		}
		if dsType.(string) == DefaultDataSource && strings.EqualFold(dsName, opts.datasource) {
			// overwrite datasource name with the one from Grafana when the names differ by case
			opts.datasource = dsName
			return true
		}
	}

	return false
}

func checkVersion(inputVersion string) bool {
	v1, err := goversion.NewVersion(inputVersion)
	if err != nil {
		fmt.Println(err)
		return false
	}

	minV, _ := goversion.NewVersion(grafanaMinVers)

	// Not using a constraint check since a pre-release version (e.g. 8.4.0-beta1) never matches
	// a constraint specified without a pre-release https://github.com/hashicorp/go-version/pull/35

	satisfies := v1.GreaterThanOrEqual(minV)
	if !satisfies {
		fmt.Printf("%s is not >= %s", v1, minV)
	}
	return satisfies
}

func checkFolder(folder *Folder) error {

	result, status, code, err := sendRequestArray(opts, "GET", "/api/folders?limit=1000", nil)

	if err != nil {
		return err
	}

	if code != 200 {
		return errors.New("server response: " + status)
	}

	if len(result) == 0 {
		return nil
	}

	for _, x := range result {
		if name, ok := x["title"]; ok {
			if name.(string) == folder.name {
				if id, idExist := x["id"]; idExist {
					folder.id = int64(id.(float64))
					if uid, uidExist := x["uid"]; uidExist {
						folder.uid = uid.(string)
					}
				}
			}
		}
	}

	return nil
}

func createServerFolder(folder *Folder) error {

	request := make(map[string]any)

	request["title"] = folder.name

	result, status, code, err := sendRequest(opts, "POST", "/api/folders", request)

	if err != nil {
		return err
	}

	if code != 200 {
		return errors.New("server response: " + status)
	}

	folder.id = int64(result["id"].(float64))
	folder.uid = result["uid"].(string)

	return nil
}

func sendRequest(opts *options, method, url string, query map[string]any) (map[string]any, string, int, error) {

	var result map[string]any

	data, status, code, err := doRequestWithRetry(opts, method, url, query, 3)
	if err != nil {
		return result, status, code, err
	}

	if err = json.Unmarshal(data, &result); err != nil {
		fmt.Printf("raw response (%d - %s):\n", code, status)
		fmt.Println(string(data))
	}
	return result, status, code, err
}

func sendRequestArray(opts *options, method, url string, query map[string]any) ([]map[string]any, string, int, error) {

	var result []map[string]any

	data, status, code, err := doRequestWithRetry(opts, method, url, query, 3)
	if err != nil {
		return result, status, code, err
	}

	if err = json.Unmarshal(data, &result); err != nil {
		fmt.Printf("raw response (%d - %s):\n", code, status)
		fmt.Println(string(data))
	}
	return result, status, code, err
}

func doRequestWithRetry(opts *options, method, url string, query map[string]any, retry int) ([]byte, string, int, error) {
	var (
		status string
		code   int
		err    error
		data   []byte
	)

	for {
		retry--
		data, status, code, err = doRequest(opts, method, url, query)
		if err != nil || code >= 500 {
			fmt.Printf("  request failed, retrying. retry=%d code=%d err=%+v url=%s\n", retry, code, err, url)
			if retry > 0 {
				continue
			}
		}
		break
	}
	return data, status, code, err
}

func doRequest(opts *options, method, url string, query map[string]any) ([]byte, string, int, error) {

	var (
		request  *http.Request
		response *http.Response
		status   string
		code     int
		err      error
		buf      *bytes.Buffer
		data     []byte
	)

	if query != nil {
		if data, err = json.Marshal(query); err != nil {
			return nil, status, code, err
		}
		buf = bytes.NewBuffer(data)
		request, err = requests.New(method, opts.addr+url, buf)
	} else {
		request, err = requests.New(method, opts.addr+url, nil)
	}

	if err != nil {
		return nil, status, code, err
	}

	// fmt.Printf("(debug) send request [%s]\n", request.URL.String())

	request.Header = opts.headers

	if response, err = opts.client.Do(request); err != nil {
		return nil, status, code, err
	}

	status = response.Status
	code = response.StatusCode

	//goland:noinspection GoUnhandledErrorResult
	defer response.Body.Close()
	data, err = io.ReadAll(response.Body)
	return data, status, code, err
}

var opts = &options{}

var Cmd = &cobra.Command{
	Use:   "grafana",
	Short: "Import/export Grafana dashboards",
	Long:  "Grafana tool - Import/Export Grafana dashboards",
}

var importCmd = &cobra.Command{
	Use:   "import",
	Short: "import Grafana dashboards",
	// Added so directory and serverfolder are required arguments except when both are empty. When both are empty use long accepted Harvest defaults
	PersistentPreRun: func(cmd *cobra.Command, _ []string) {
		dir := cmd.Flags().Lookup("directory")
		folder := cmd.Flags().Lookup("serverfolder")
		if dir.Value.String() == "" && folder.Value.String() == "" {
			dir.Changed = true
			folder.Changed = true
		}
	},
	Run: doImport,
	Example: `
# Add the default set of cDot and 7mode dashboards from local directory grafana/dashboards to my.grafana.server 
grafana import --addr my.grafana.server:3000

# Add the dashboards from the local directory to the server_folder on my.grafana.server
grafana import --addr my.grafana.server:3000 --directory [local] --serverfolder [server_folder]`,
}

var exportCmd = &cobra.Command{
	Use:   "export",
	Short: "export Grafana dashboards",
	Run:   doExport,
	Example: `
# Export all the dashboards contained in the server_folder on my.grafana.server and write them to the local directory
grafana export --addr my.grafana.server:3000 --serverfolder server_folder --directory local`,
}

var customizeCmd = &cobra.Command{
	Use:   "customize",
	Short: "customize Grafana dashboards and write to filesystem",
	Run:   doCustomize,
	Example: `
# Customize all the dashboards recursively contained in grafana/dashboards and write them to ~/harvest-dashboards.
grafana customize --directory grafana/dashboards --output-dir ~/harvest-dashboards --prefix netapp_ --datasource my_datasource`,
}

func init() {
	Cmd.AddCommand(importCmd, exportCmd, customizeCmd, metricsCmd)
	addCommonFlags(importCmd, exportCmd, customizeCmd)
	addImportExportFlags(importCmd, exportCmd)
	addCustomizeFlags(customizeCmd)

	importCmd.PersistentFlags().StringSliceVar(&opts.labels, "labels", nil,
		"For each label, create a variable and add as chained query to other variables")
	importCmd.PersistentFlags().BoolVar(&opts.addMultiSelect, "multi", true,
		"Modify the dashboards to add multi-select dropdowns for each variable")
	importCmd.PersistentFlags().BoolVar(&opts.forceImport, "force", false,
		"Import even if the datasource name is not defined in Grafana")
	_ = importCmd.PersistentFlags().MarkHidden("multi")
	_ = importCmd.PersistentFlags().MarkHidden("force")

	metricsCmd.PersistentFlags().StringVarP(&opts.dir, "directory", "d",
		"", "local directory that contains dashboards (searched recursively).")
}

func addCustomizeFlags(cmd *cobra.Command) {
	cmd.PersistentFlags().StringVarP(&opts.customizeDir, "output-dir", "o", "", "Write customized dashboards to the local directory. The directory must not exist")
}

func addCommonFlags(commands ...*cobra.Command) {
	for _, cmd := range commands {
		cmd.PersistentFlags().StringVar(&opts.config, "config", "./harvest.yml", "harvest config file path")
		cmd.PersistentFlags().StringVar(&opts.svmRegex, "svm-variable-regex", "", "SVM variable regex to filter SVM query results")
		cmd.PersistentFlags().StringVarP(&opts.prefix, "prefix", "p", "", "Use global metric prefix in queries")
		cmd.PersistentFlags().StringVarP(&opts.datasource, "datasource", "s", DefaultDataSource, "Name of your Prometheus datasource used by the imported dashboards")
		cmd.PersistentFlags().BoolVarP(&opts.variable, "variable", "v", false, "Use datasource as variable, overrides: --datasource")
		cmd.PersistentFlags().StringVarP(&opts.dir, "directory", "d", "", "When importing, import dashboards from this local directory.\nWhen exporting, local directory to write dashboards to")

		_ = cmd.PersistentFlags().MarkHidden("svm-variable-regex")
		_ = cmd.MarkPersistentFlagRequired("directory")
	}
}

func addImportExportFlags(commands ...*cobra.Command) {
	for _, cmd := range commands {
		cmd.PersistentFlags().StringVarP(&opts.addr, "addr", "a", "http://127.0.0.1:3000", "Address of Grafana server (IP, FQDN or hostname)")
		cmd.PersistentFlags().StringVarP(&opts.token, "token", "t", "", "API token issued by Grafana server for authentication")
		cmd.PersistentFlags().BoolVarP(&opts.useHTTPS, "https", "S", false, "Use HTTPS")
		cmd.PersistentFlags().BoolVarP(&opts.overwrite, "overwrite", "o", false, "Overwrite existing dashboard with same title")
		cmd.PersistentFlags().BoolVarP(&opts.useInsecureTLS, "insecure", "k", false, "Allow insecure server connections when using SSL")
		cmd.PersistentFlags().StringVarP(&opts.serverfolder.name, "serverfolder", "f", "", "Grafana folder name for dashboards")

		_ = cmd.MarkPersistentFlagRequired("serverfolder")
	}
}
